import json
from io import StringIO
from lura.attrs import ottr
from lura.formats import base

class Encoder(json.JSONEncoder):

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def default(self, item):
    if isinstance(item, set):
      return list(item)
    elif hasattr(item, '__str__'):
      return str(item)
    else:
      return repr(item)

class Format(base.Format):
  '''
  This class loads json using ordered dictionaries, and `repr()`s unsupported
  types on dump rather than raising.
  '''

  object_pairs_hook = ottr

  def __init__(self):
    super().__init__()

  def loads(self, data, **kwargs):
    kwargs.setdefault('object_pairs_hook', self.object_pairs_hook)
    return json.loads(data, **kwargs)

  def loadf(self, src, encoding=None, **kwargs):
    with open(src, encoding=encoding) as fd:
      return self.loadfd(fd, **kwargs)

  def loadfd(self, fd, **kwargs):
    kwargs.setdefault('object_pairs_hook', self.object_pairs_hook)
    return json.load(fd, **kwargs)

  def dumps(self, data, **kwargs):
    kwargs.setdefault('cls', Encoder)
    return json.dumps(data, **kwargs)

  def dumpf(self, data, dst, encoding=None, **kwargs):
    with open(dst, 'w', encoding=encoding) as fd:
      self.dumpfd(data, fd, **kwargs)

  def dumpfd(self, data, fd, **kwargs):
    kwargs.setdefault('cls', Encoder)
    json.dump(data, fd, **kwargs)
    if hasattr(fd, 'flush') and callable(fd.flush):
      fd.flush()



import sys
import pkg_resources

class Assets:

  @classmethod
  def join(cls, *args):
    return '/'.join(arg.rstrip('/') for arg in args)

  @classmethod
  def split(cls, path):
    return path.strip('/').split('/')

  @classmethod
  def basename(cls, path):
    return cls.split(path)[-1]

  @classmethod
  def dirname(cls, path):
    return cls.join(cls.split(path)[:-1])

  def __init__(self, package, prefix=None):
    super().__init__()
    self.package = package
    self.prefix = prefix

  def path(self, path):
    if self.prefix:
      return self.join(self.prefix, path)
    return path

  def load(self, path):
    return pkg_resources.resource_string(self.package, self.path(path))

  def loads(self, path, encoding=None):
    encoding = encoding or sys.getdefaultencoding()
    buf = pkg_resources.resource_string(self.package, self.path(path))
    return buf.decode(encoding)

  def copy(self, src, dst):
    buf = self.load(src)
    with open(dst, 'wb') as file:
      file.write(buf)

  def open(self, path):
    return pkg_resources.resource_stream(self.package, self.path(path))

  def list(self, path, long=False):
    files = pkg_resources.resource_listdir(self.package, self.path(path))
    if long:
      files = [self.join(path, file) for file in files]
    return files

  def exists(self, path):
    return pkg_resources.resource_exists(self.package, self.path(path))

  def isdir(self, path):
    return pkg_resources.resource_isdir(self.package, self.path(path))

  def isfile(self, path):
    return self.exists(path) and not self.isdir(path)

  def print(self, path, *args, **kwargs):
    print(self.loads(path), *args, **kwargs)

  def bind(self, path):
    prefix = self.join(self.prefix, path)
    return type(self)(package=self.package, prefix=prefix)



import types
from collections import MutableMapping
from collections import MutableSequence
from lura.attrs import attr

# FIXME this needs a different name
def deepcopy(obj, map_cls=attr, seq_cls=None):
  types = (MutableMapping, MutableSequence)
  if isinstance(obj, MutableMapping):
    cls = map_cls or type(obj)
    return cls(
      (k, deepcopy(v, map_cls, seq_cls))
      if isinstance(v, types)
      else (k, v)
      for (k, v) in obj.items()
    )
  elif isinstance(obj, MutableSequence):
    cls = seq_cls or type(obj)
    return cls(
      deepcopy(item, map_cls, seq_cls)
      if isinstance(item, types)
      else item
      for item in obj
    )
  else:
    raise ValueError(f'obj is not a MutableMapping or MutableSequence: {obj}')

class Kwargs:

  def __init__(self, *args, **kwargs):
    super().__init__()
    vars(self).update(kwargs)

def isexc(o):
  '''
  `True` if `o` is a tuple as returned by `sys.exc_info()`, else
  `False`.
  '''

  return isinstance(o, tuple) and len(o) == 3 and (
    isinstance(o[0], type) and
    isinstance(o[1], o[0]) and
    isinstance(o[2], types.TracebackType)
  )



'SSL RPC services using RPyC.'

import logging
import rpyc
from rpyc.core.service import SlaveService
from rpyc.utils.authenticators import SSLAuthenticator
from rpyc.utils.server import ThreadedServer

logger = logging.getLogger(__name__)

Service = SlaveService

def listen(service, host, port, key_path, cert_path, sync_timeout, backlog):

  name = service.get_service_name()
  protocol_config = dict(
    allow_all_attrs = True,
    allow_delattr = True,
    allow_public_attrs = True,
    allow_setattr = True,
    logger = logger,
    sync_request_timeout = sync_timeout)
  authenticator = SSLAuthenticator(key_path, cert_path)
  server = ThreadedServer(service,
    hostname=host, port=port, protocol_config=protocol_config,
    authenticator=authenticator, backlog=backlog, logger=logger)
  server.start()

def _patch_close(conn, on_close):
  conn_close = conn.close
  def close(*args, **kwargs):
    on_close()
    conn_close(*args, **kwargs)
    conn.close = conn_close
  conn.close = close

def connect(
  host, port, key_path, cert_path, sync_timeout, on_connect=None,
  on_close=None
):
  protocol_config = dict(
    allow_all_attrs = True,
    allow_delattr = True,
    allow_public_attrs = True,
    allow_setattr = True,
    logger = logger,
    sync_request_timeout = sync_timeout)
  conn = rpyc.ssl_connect(
    host, port=port, keyfile=key_path, certfile=cert_path,
    config=protocol_config)
  if on_connect:
    on_connect()
  name = conn.root.get_service_name()
  if on_close:
    _patch_close(conn, on_close)
  conn.host = host
  conn.port = port
  conn.service = conn.root
  return conn



from lura.formats import csv
from lura.formats import json
from lura.formats import pickle
from lura.formats import yaml

csv = csv.Format()
json = json.Format()
pickle = pickle.Format()
yaml = yaml.Format()

exts = dict(
  csv = csv,
  jsn = json,
  json = json,
  pickle = pickle,
  pckl = pickle,
  yaml = yaml,
  yml = yaml,
)

def for_ext(ext):
  if ext not in exts:
    raise ValueError(f'No format for file extension: {ext}')
  return exts[ext]

def for_path(path):
  ext = path.rsplit('.', 1)[-1]
  if ext not in exts:
    raise ValueError(f'No format for file extension: {path}')
  return exts[ext]



import logging
import logging.config
import os
import sys
import time
import traceback
import yaml
from collections import defaultdict
from io import StringIO
from lura.attrs import attr

#####
## handy values

# log format presets (these depend on ExtraInfoFilter)
formats = attr(
  bare    = '%(message)s',
  runtime = '%(x_runtime)-12.3f %(message)s',
  user    = '%(x_runtime)-8.3f %(x_char)s %(message)s',
  hax     = '%(x_runtime)-8.3f %(x_modules)20s %(x_char)s %(message)s',
  daemon  = '%(asctime)s %(x_module)10s %(x_char)s %(message)s',
  verbose = '%(asctime)s %(x_runtime)12.3f %(name)s %(x_char)s %(message)s',
)

logging.formats = formats

default_datefmt = '%Y-%m-%d %H:%M:%S'

#####
## utility classes

class ExtraInfoFilter(logging.Filter):
  '''
  Provides additional fields to log records:

  - modules - 'parent_module.calling_module'
  - module - 'calling_module'
  - runtime - number of seconds this class has been loaded as float
  - char - a single character uniquely identifying the log level
  '''

  initialized = time.time()
  default_char = ' '
  name_to_char = defaultdict(
    lambda: default_char,
    DEBUG    = '+',
    INFO     = '|',
    WARNING  = '>',
    ERROR    = '*',
    CRITICAL = '!',
  )

  def filter(self, record):
    modules = record.name.split('.')
    record.x_modules = '.'.join(modules[-2:])
    record.x_module = modules[-1]
    record.x_char = self.name_to_char.get(record.levelname)
    record.x_runtime = time.time() - self.initialized
    return True

class MultiLineFormatter(logging.Formatter):
  '''
  Format messages containing lineseps as though each line were a log
  message.
  '''

  def format(self, record):
    if not isinstance(record.msg, str) or os.linesep not in record.msg:
      record.message = super().format(record)
      return record.message
    msg = record.msg
    with StringIO() as buf:
      for line in record.msg.split(os.linesep):
        record.msg = line
        buf.write(super().format(record) + os.linesep)
      record.message = buf.getvalue().rpartition(os.linesep)[0]
    record.msg = msg
    return record.message

class Logger(logging.getLoggerClass()):
  '''
  Logger subclass with the following changes:

  - __getitem__(log_level) -> callable log method for log_level
  - setConsoleFormat(format, datefmt) will set the format for any stream
    handler writing to stdout or stderr
  - log level constants are set on the class, and will be updated by
    logutils.add_level()
  - exceptions are sent through formatters rather than being printed
    bare
  '''

  def __getitem__(self, level):
    'Return the log method for the given log level number.'

    name = logging._levelToName[level].lower()
    return getattr(self, name)

  def setConsoleFormat(self, format, datefmt=None):
    'Set the output format on any handler for stdout or stderr.'

    datefmt = datefmt or default_datefmt
    formatter = MultiLineFormatter(format, datefmt)
    std = (sys.stdout, sys.stderr)
    for handler in self.handlers:
      if hasattr(handler, 'stream') and handler.stream in std:
        handler.setFormatter(formatter)

  def _log(self, level, msg, *args, exc_info=None, **kwargs):
    if exc_info:
      if isinstance(exc_info, BaseException):
        exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
      else:
        exc_info = sys.exc_info()
      tb = ''.join(traceback.format_exception(*exc_info)).rstrip()
      super()._log(level, msg, *args, **kwargs)
      super()._log(level, tb, *args, **kwargs)
    else:
      super()._log(level, msg, *args, **kwargs)

for level, name in logging._levelToName.items():
  setattr(Logger, name, level)

logging.setLoggerClass(Logger)

#####
## utilities

def number_for_name(name):
  return logging._nameToLevel[name]

def name_for_number(number):
  return logging._levelToName[number]

def build_log_method(number):
  def log_level(self, *args, **kwargs):
    if self.isEnabledFor(number):
      self._log(number, msg, args, **kwargs)
  return log_level

def add_level(self, name, number, char=None):
  '''
  Add a log level.

  - Sets the attribute `name` to `number` on module `logging`
  - Sets the attribute `name` to `number` on type `logging.Logging`
  - Creates a log method for the level and sets it on `logging.Logging` as
    attribute `name.lower()`
  - Sets `char` for `name` on `ExtraInfoFilter` when provided
  '''

  if name in logging._nameToLevel:
    raise ValueError(f'Level name {name} already in use')
  if number in logging._levelToName:
    raise ValueError(f'Level number {number} already in use')
  if char is not None:
    if char in ExtraInfoFilter.name_to_char.values():
      raise ValueError(f'Level char {char} already in use')
  logging.addLevelName(number, name)
  setattr(logging, name, number)
  setattr(Logger, name, number)
  setattr(Logger, name.lower(), build_log_method(number))
  if char is not None:
    ExtraInfoFilter.name_to_char[name] = char

#####
## logging configurator

class Configurator:
  '''
  Build and apply a dictConfig using features of `lura.logutils`.

  - Creates a filter `extra_info` as `ExtraInfoFilter`
  - Creates a formatter `multiline` as `MultiLineFormatter` using `format`
    and `datefmt`
  - Creates a handler 'stderr` as `StreamHandler` for `sys.stderr` using the
    `extra_info` filter and the `multiline` formatter
  - Sets the `level` and `stderr` handler for a package's root logger
  '''

  def __init__(self, package, format, datefmt, level=logging.INFO):
    super().__init__()
    self.package = package
    self.format = format
    self.datefmt = datefmt or default_datefmt
    self.level = level

  def yaml(self, str):
    return yaml.safe_load(str)

  @property
  def filters(self):
    filters = self.yaml('''
      extra_info: {}
    ''')
    filters['extra_info']['()'] = ExtraInfoFilter
    return filters

  @property
  def formatters(self):
    formatters = self.yaml(f'''
      multiline:
        format: '{self.format}'
        datefmt: '{self.datefmt}'
    ''')
    formatters['multiline']['()'] = MultiLineFormatter
    return formatters

  @property
  def handlers(self):
    return self.yaml('''
      stderr:
        class: logging.StreamHandler
        stream: ext://sys.stderr
        filters: [extra_info]
        formatter: multiline
    ''')

  @property
  def loggers(self):
    return self.yaml(f'''
      {self.package}:
        handlers: [stderr]
        level: {name_for_number(self.level)}
    ''')

  @property
  def config(self):
    config = self.yaml('''
      version: 1
      disable_existing_loggers: false
    ''')
    config['filters'] = self.filters
    config['formatters'] = self.formatters
    config['handlers'] = self.handlers
    config['loggers'] = self.loggers
    return config

  def configure(self):
    logging.config.dictConfig(self.config)

def configure(package, format, datefmt=None, level=logging.INFO):
  Configurator(package, format, datefmt, level=level).configure()



import ctypes
import inspect
import logging
import sys
import threading
from lura.attrs import attr
from lura.time import poll
from time import sleep

log = logging.getLogger(__name__)

class Synchronize:
  '''
  As a decorator for functions or instance, class, or static methods:

  - functions will lock on the function
  - instance methods will lock on the instance
  - class methods will lock on the class
  - static methods will lock on the class

  As a context manager:

  The object to lock on is passed to the constructor as `wrapped`.

  Inspired by:

  http://blog.dscpl.com.au/2014/01/the-missing-synchronized-decorator.html
  '''

  lock = threading.RLock()
  lock_name = '__synchronize__'

  def __init__(self, wrapped=None, owner=None, type=threading.RLock):
    super().__init__()
    self.wrapped = wrapped
    self.owner = owner
    self.lock_type = type

  def get_lock(self, obj):
    if self.owner is not None:
      obj = self.owner
    ctx = vars(obj)
    if self.lock_name not in ctx:
      with self.lock:
        if self.lock_name not in ctx:
          setattr(obj, self.lock_name, self.lock_type())
    return ctx[self.lock_name]

  def __enter__(self):
    lock = self.get_lock(self.wrapped)
    lock.acquire()

  def __exit__(self, *exc_info):
    lock = self.get_lock(self.wrapped)
    lock.release()

  def __get__(self, obj, cls=None):
    # instance, class, and static methods will enter here
    if obj and cls:
      def wrapper_instance(*args, **kwargs):
        with self.get_lock(obj):
          return self.wrapped(obj, *args, **kwargs)
      return wrapper_instance
    elif cls:
      wrapped = self.wrapped.__get__(None, cls)
      def wrapper_class(*args, **kwargs):
        with self.get_lock(cls):
          return wrapped(*args, **kwargs)
      return wrapper_class
    else:
      assert(False)

  def __call__(self, *args, **kwargs):
    # functions will enter here
    with self.get_lock(self.wrapped):
      return self.wrapped(*args, **kwargs)

synchronize = Synchronize

def _async_raise(tid, exc_type):
  '''
  Raise an exception in thread with id `tid`. Return `True` if the exception
  was sent to the thread, or `False` if thead id `tid` was not found.
  '''

  if not inspect.isclass(exc_type):
    raise TypeError(f'exc_type is not a class: {exc_type}')
  res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
    ctypes.c_long(tid), ctypes.py_object(exc_type))
  return res != 0

class Cancelled(RuntimeError):

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

class Thread(threading.Thread):

  cancel_poll_interval = 0.05

  @classmethod
  def spawn(cls, *args, **kwargs):
    thread = cls(*args, **kwargs)
    thread.start()
    return thread

  def __init__(
    self, group=None, target=None, name=None, args=(), kwargs={}, *,
    daemon=None, reraise=False
  ):
    super().__init__(
      group=group, name=name, daemon=daemon)
    self._thread_target = target
    self._thread_args = args
    self._thread_kwargs = kwargs
    self._thread_reraise = reraise
    self._thread_result = None
    self._thread_error = None

  def _thread_work(self):
    try:
      self._thread_result = self._thread_target(
        *self._thread_args, **self._thread_kwargs)
    except Exception:
      self._thread_error = sys.exc_info()
      if self._thread_reraise:
        raise

  def _thread_get_id(self):
    if not self.isAlive():
      return None
    if hasattr(self, '_thread_id'):
      return self._thread_id
    for tid, tobj in threading._active.items():
      if tobj is self:
        return tid
    return None

  def cancel(self, timeout=None, exc_type=Cancelled, force=False):
    '''
    Cancel by raising an exception in the running thread. This approach is
    not recommended as it is not reliable and can fail in surprising ways;
    however, it can work well when designed for. Two things to note:

    - If the thread is executing C code (`sleep()`, `socket.accept()`), the
      exception will not be raised until execution returns to the python
      interpreter.

    - Any C code running in the thread which invokes `PyErr_Clear` from the
      C api before returning will effectively clear the raised exception.

    When `force` is `False`, one exception will be raised in the thread. When
    `force` is `True`, the thread will be barraged with exceptions until it
    stops, or the timeout expires.

    In the former case, the thread will have an opportunity to handle (or
    discard) the raised exception before returning. In the latter case,
    any exception handlers will themselves be interrupted by the barrage of
    exceptions.

    See the original implementation notes for more details:

    http://tomerfiliba.com/recipes/Thread2/
    '''

    def cancel(tid):
      _async_raise(tid, exc_type)
      sleep(0)
      return not self.isAlive()

    tid = self._thread_get_id()
    if tid is None:
      if not self.isAlive():
        return True
      self.join(timeout)
      return self.isAlive()
    if not force:
      cancel(tid)
      self.join(timeout)
      return self.isAlive()
    timeout = -1 if timeout is None else timeout
    test = lambda: cancel(tid)
    return poll(test, timeout=timeout, pause=self.cancel_poll_interval)

  def run(self):
    '''
    If no target is passed to the constructor, this method is used as the
    target.
    '''

    raise NotImplementedError()

  def start(self):
    # FIXME !!!
    if not self._thread_target:
      self._thread_target = self.run
    self.run = self._thread_work
    super().start()

  @property
  def result(self):
    return self._thread_result

  @property
  def error(self):
    return self._thread_error



import os
import re
import subprocess
from distutils.util import strtobool

shjoin = subprocess.list2cmdline

def prefix(string, prefix, linesep=os.linesep):
  'Return `string` with each line prefixed with `prefix`.'

  return linesep.join(f'{prefix}{line}' for line in string.split(linesep))

def as_bool(val):
  'Use `strtobool` to parse `str`s into `bool`s.'

  if val == '':
    return False
  return bool(strtobool(val)) if isinstance(val, str) else bool(val)

def camel_to_snake(string):
  return re.sub('(?!^)([A-Z]+)', r'_\1', string).lower()



'Periodic task scheduler.'

import logging
import schedule
from abc import abstractmethod
from lura import threads
from time import sleep

logger = logging.getLogger(__name__)

class Task(threads.Thread):
  '''
  This is a base class for scheduled tasks, which are ephemeral threads
  spawned at the appropriate time and then discarded by the `Scheduler`.

  Subclasses shall implement `work()` to perform the scheduled task.
  '''

  log_level = logger.INFO

  def __init__(self, scheduler, *args, **kwargs):
    for _ in kwargs:
      # don't allow these to be forwarded to the Thread ctor
      assert(_ not in ('target', 'args', 'kwargs'))
    super().__init__(*args, **kwargs)
    self._scheduler = scheduler

  def _on_error(self):
    '''
    This will be called from inside an `except` block (so you can call
    `sys.exc_info()`) when unhandled exceptions are encountered.
    '''

    pass

  @abstractmethod
  def work(self):
    pass

  def run(self):
    log = logger[self.log_level]
    try:
      log(f'{type(self).__name__} starting')
      self.work()
    except Exception:
      log(f'Unhandled exception in {type(self).__name__}', exc_info=True)
      self._on_error()
    finally:
      log(f'{type(self).__name__} stopping')

class Scheduler:
  '''
  Periodic task scheduler implemented using the `schedule` package.

  Pass the constructor a `schedule`, which is a list of pairs:

  `[(schedule.Job, lura.at.Task), ...]`

  Example:
  ```
  import schedule
  from lura import at

  class CycleLogFiles(at.Task): pass
  class StashLogFIles(at.Task): pass

  sched = [
    (schedule.every().day.at('00:15'), CycleLogFiles),
    (schedule.every().day.at('00:30'), StashLogFiles),
  ]

  scheduler = at.Scheduler(sched)
  scheduler.start()
  ```
  '''

  log_level = logger.INFO

  def __init__(self, schedule):
    super().__init__()
    self._schedule = schedule
    self._applied = False
    self._working = False

  def _on_error(self):
    '''
    This will be called from inside an `except` block (so you can call
    `sys.exc_info()`) when unhandled exceptions are encountered.
    '''

    pass

  def _format_task(self, job, task):
    if job.at_time is not None:
      return '%s every %s %s at %s' % (
        task.__name__,
        job.interval,
        job.unit[:-1] if job.interval == 1 else job.unit,
        job.at_time,
      )
    else:
      fmt = (
        '%(task)s every %(interval)s ' +
        'to %(latest)s ' if job.latest is not None else '' +
        '%(unit)s'
      )
      return fmt % dict(
        task = task.__name__,
        interval = job.interval,
        latest = job.latest,
        unit = job.unit[:-1] if job.interval == 1 else job.unit,
      )

  def _apply_schedule(self):
    log = logger[self.log_level]
    if self._applied:
      log('Asked to apply schedule when already applied')
      return
    for job, task in self._schedule:
      job.do(task.spawn, self)
      log('Scheduled ' + self._format_task(job, task))
    self._applied = True

  def run(self):
    self._working = True
    log = logger[self.log_level]
    log('Task scheduler starting')
    self._apply_schedule()
    try:
      import schedule
      while self._working:
        schedule.run_pending()
        sleep(1)
    except Exception:
      log('Unhanlded exception in task scheduler', exc_info=True)
      self._on_error()
      raise
    finally:
      log('Task scheduler stopping')

  start = run

  def stop(self):
    self._working = False



import io
import ipaddress
import requests
import socket
import sys

def resolve(hostname):
  try:
    info = socket.getaddrinfo(hostname, 80, proto=socket.IPPROTO_TCP)
    return info[0][4][0]
  except socket.gaierror:
    return None

def is_ip_address(string):
  try:
    ipaddress.ip_address(string)
    return True
  except ValueError:
    return False

def wgetfd(url, fd, bufsize=64*1024):
  with requests.get(url, stream=True) as req:
    req.raise_for_status()
    for buf in req.iter_content(chunk_size=bufsize):
      if buf:
        fd.write(buf)

def wgetf(url, path, bufsize=64*1024):
  with open(path, 'wb') as pathf:
    self.wgetfd(url, pathf, bufsize)

def wget(url):
  with io.BytesIO() as buf:
    wgetfd(url, buf)
    return buf.getvalue()

def wgets(url, encoding=None):
  encoding = encoding or sys.getdefaultencoding()
  return wget(url).decode(encoding)



import click

class StartsWithGroup(click.Group):

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def get_command(self, ctx, name):
    res = super().get_command(ctx, name)
    if res is not None:
      return res
    names = [_ for _ in self.list_commands(ctx) if _.startswith(name)]
    if not names:
      return None
    elif len(names) == 1:
      return super().get_command(ctx, names[0])
    names = ', '.join(sorted(names))
    ctx.fail(f'Too many matches: {names}')



'''
Send messages to e.g. ms-teams or discord.

This is a simple api that can accommodate multiple messaging services.

There's also `Messenger` class that will queue outgoing and send messages
on a schedule as to not trigger flood protection on your webhooks.
'''

import logging
import queue
import requests
from lura import threads
from lura.attrs import attr
from lura.formats import json
from lura.time import poll
from time import sleep

logger = logging.getLogger(__name__)

def teams(webhook, title='', subtitle='', summary='', fields={}, timeout=20.0):
  'Send a message to MS Teams.'

  payload = {
    '@type': 'MessageCard',
    '@context': 'http://schema.org/extensions',
    'summary': summary, # XXX what does this actually do?
    'sections': [
      {
        'activityTitle': title,
        'activitySubtitle': subtitle,
        'facts': [{'name': n, 'value': v} for (n, v) in fields.items()]
      }
    ]
  }
  headers = {'Content-Type': 'application/json'}
  res = requests.post(
    webhook, headers=headers, data=json.dumps(payload), timeout=timeout)
  try:
    res.raise_for_status()
    return None
  except requests.exceptions.HTTPError:
    return res

def discord(
  webhook, title=None, subtitle=None, summary=None, fields={}, timeout=20.0
):
  'Send a message to Discord.'

  embed = attr()
  if title:
    embed.title = title
  if subtitle:
    embed.description = subtitle
  if fields:
    fields = [{'name': n, 'value': v} for (n, v) in fields.items()]
    embed.fields = fields
  if summary:
    embed.footer = {'text': summary}
  payload = {'embeds': [embed]}
  headers = {'Content-Type': 'application/json'}
  res = requests.post(
    webhook, headers=headers, data=json.dumps(payload), timeout=timeout)
  try:
    res.raise_for_status()
    return None
  except requests.exceptions.HTTPError:
    return res

class Messenger:
  '''
  This class queues messages for teams and discord and sends them a regular
  interval so as to not trigger flood protection. Create an instance and call
  `start()` to execute the main loop.
  '''

  log_level = logger.INFO

  _senders = {'teams': teams, 'discord': discord}

  # how long will we block in queue.get() before giving up
  _queue_get_timeout = 5

  def __init__(self, teams_webhook=None, discord_webhook=None, pulse=4.0):
    super().__init__()
    self._webhooks = {
      'teams': teams_webhook,
      'discord': discord_webhook,
    }
    self._pulse = pulse
    self._queue = queue.Queue()
    self._working = False

  def is_idle(self):
    return self._queue.empty()

  idle = property(is_idle)

  def wait_for_idle(self, timeout):
    return poll(self.is_idle, timeout=timeout, pause=1)

  def send(self, **kwargs):
    self._queue.put(kwargs)

  def _send(self, kwargs):
    'Send the message with all configured services.'

    log = logger[self.log_level]
    sent = False
    for service in 'teams', 'discord':
      webhook = self._webhooks[service]
      if not webhook:
        continue
      send = self._senders[service]
      try:
        res = send(webhook, **kwargs)
        if res is None:
          sent = True
      except Exception:
        log(f'Unhandled exception for {service} message', exc_info=True)
    return sent

  def _work(self):
    'Get a set of kwargs from the queue and try to send them.'

    try:
      kwargs = self._queue.get(block=True, timeout=self._queue_get_timeout)
    except queue.Empty:
      return
    try:
      sent = self._send(kwargs)
    finally:
      self._queue.task_done()
    if sent:
      sleep(self._pulse)

  def _loop(self):
    'Run the work loop.'

    while self._working:
      self._work()

  def run(self):
    'Main entry point for the messaging work loop.'

    log = logger[self.log_level]
    self._working = True
    log('Messenger starting')
    try:
      self._loop()
    except Exception:
      log(f'Unhandled exception in messenger', exc_info=True)
      raise
    finally:
      log('Messenger stopping')

  start = run

  def stop(self):
    'Break the work loop.'

    self._working = False



import io
import logging
import os
from abc import abstractmethod
from lura import fs
from lura.formats import json
from lura.run import run
from lura.time import poll

log = logging.getLogger(__name__)

def convert_opts(opts):
  def convert_opt(name, value):
    name = name.replace('_', '-')
    if value in (None, ):
      raise ValueError(f"Invalid value for kubectl argument {name}: {value}")
    elif isinstance(value, bool):
      value = 'true' if value else 'false'
    return f"--{name}={value}"
  return [convert_opt(name, value) for name, value in opts.items()]

def kubectl(cmd, args=[], opts={}, cwd=None, enforce=True):
  argv = [kubectl.bin, cmd]
  argv.extend(args)
  argv.extend(convert_opts(opts))
  with run.log(log, kubectl.log_level):
    return run(argv, cwd=cwd, enforce=enforce)

kubectl.bin = 'kubectl'
kubectl.log_level = log.DEBUG

def get(*args, **opts):
  if 'output' in opts:
    return kubectl('get', args, opts).stdout
  else:
    opts['output'] = 'json'
    return json.loads(kubectl('get', args, opts).stdout)

def describe(*args, **opts):
  return kubectl('describe', args, opts).stdout

def apply(*args, **opts):
  kubectl('apply', args, opts)

def delete(*args, enforce=True, **opts):
  kubectl('delete', args, opts, enforce=enforce)

def logs(*args, **opts):
  return kubectl('logs', args, opts).stdout

class ResourceFiles(fs.TempDir):

  def __init__(self, resources, keep=False):
    super().__init__(prefix='lura-kube.', keep=keep)
    self.resources = resources

  def __enter__(self):
    log.debug(f'Creating resource files')
    temp_dir = super().__enter__()
    files = []
    for name, resource in self.resources:
      dst = os.path.join(temp_dir, f'{name}.yaml')
      log.debug(f'    {dst}')
      fs.dumps(dst, resource)
      files.append(dst)
    del self.resources
    return files

class Application:

  delete_timeout = 60.0

  def __init__(self, name):
    super().__init__()
    self.name = name

  @abstractmethod
  def _get_resources(self):
    return []

  @abstractmethod
  def _get_pods(self):
    return []

  def get_resources(self):
    return self._get_resources()

  def get_pods(self):
    return self._get_pods()

  def get_manifest(self):
    with io.StringIO() as buf:
      for _, resource in self._get_resources():
        buf.write('---')
        buf.write(os.linesep)
        buf.write(resource.rstrip())
        buf.write(os.linesep)
      return buf.getvalue()

  def is_applied(self):
    return bool(self._get_pods())

  def apply(self):
    log.info(f'Applying {self.name}')
    with ResourceFiles(self._get_resources()) as files:
      for file in files:
        apply(filename=file)

  def delete(self, enforce=True):
    log.info(f'Deleting {self.name}')
    with ResourceFiles(self._get_resources()) as files:
      for file in reversed(files):
        delete(filename=file, enforce=enforce)
    timeout = self.delete_timeout
    if not poll(lambda: not self.is_applied(), timeout=timeout, pause=1):
      raise TimeoutError(f'Application did not stop within {timeout} seconds')

  def logs(self, count):
    pods = self._get_pods()
    if not pods:
      return ''
    with io.StringIO() as buf:
      for pod in pods:
        name = pod.metadata.name
        lines = logs(name, tail=count)
        buf.write(f'----- {name} -----')
        buf.write(os.linesep)
        if lines:
          buf.write(lines.rstrip())
          buf.write(os.linesep)
        buf.write(os.linesep)
      return buf.getvalue()

  manifest = property(get_manifest)
  pods = property(get_pods)
  applied = property(is_applied)
  resources = property(get_resources)



import logging
import os
import shutil
import subprocess as subp
import sys
from lura.plates import jinja2
from lura.time import Timer
from shlex import quote
from tempfile import mkdtemp
from time import sleep

log = logging.getLogger(__name__)

shjoin = subp.list2cmdline

def shell_path():
  proc = subp.Popen(
    'echo $SHELL', shell=True, stdout=subp.PIPE,
    encoding=sys.getdefaultencoding())
  with proc:
    proc.wait()
    return proc.stdout.read().rstrip()

class SudoTimeout(TimeoutError):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, *kwargs)

class SudoHelper:
  '''
  A helper for implementing `popen()`-like or `run()`-like interfaces that
  support invoking commands with sudo. See the `popen()` function in this
  module for a usage example.
  '''

  shell = shell_path()
  timeout = 7.0
  sleep_interval = 0.1

  def __init__(self, password, user=None, group=None, login=None):
    super().__init__()
    self.user = user
    self.group = group
    self.password = password
    self.login = login
    self.timer = None
    self.temp_dir = None
    self.ok_path = None
    self.fifo_path = None
    self.askpass_path = None
    self.fifo = None

  @property
  def _timed_out(self):
    return self.timer.time > self.timeout

  def _update_argv(self, argv, cwd):
    argv_is_str = True
    if not isinstance(argv, str):
      argv_is_str = False
      argv = shjoin(argv)
    user_argv = argv
    argv = ['sudo']
    if self.password:
      argv.append('-A')
    if self.user:
      argv.extend(('-u', self.user))
    if self.group:
      argv.extend(('-g', self.group))
    if self.login:
      argv.append('-i')
    argv.extend((self.shell, '-c'))
    shell_argv = f'touch {quote(self.ok_path)} && '
    if cwd:
      shell_argv += f'cd {quote(cwd)} && '
    shell_argv += user_argv
    argv.append(shell_argv)
    return shjoin(argv) if argv_is_str else argv

  def _askpass_create(self):
    env = self.__dict__.copy()
    env['shell'] = self.shell
    env['quote'] = quote
    jinja2.expandsf(env, self._askpass_tmpl, self.askpass_path)
    os.chmod(self.askpass_path, 0o700)

  def _check_for_ok(self):
    return os.path.isfile(self.ok_path)

  def _fifo_create(self):
    os.mkfifo(self.fifo_path)
    os.chmod(self.fifo_path, 0o600)

  def _fifo_open(self):
    while not self._timed_out:
      if self._check_for_ok():
        return
      try:
        self.fifo = os.open(self.fifo_path, os.O_NONBLOCK | os.O_WRONLY)
        return
      except OSError:
        pass
      sleep(self.sleep_interval)
    else:
      raise SudoTimeout('Timeout opening sudo fifo')

  def _fifo_write(self):
    if self._check_for_ok():
      return
    password = self.password.encode()
    pos = 0
    end = len(password)
    while not self._timed_out:
      try:
        pos += os.write(self.fifo, password[pos:])
        if pos == end:
          return
        continue
      except BlockingIOError:
        pass
      sleep(self.sleep_interval)
    else:
      raise SudoTimeout('Timeout writing sudo fifo')

  def _fifo_close(self):
    if self.fifo is not None:
      os.close(self.fifo)
      self.fifo = None

  def _handle_fifo(self):
    self._fifo_open()
    try:
      self._fifo_write()
    finally:
      self._fifo_close()

  def _await_ok(self):
    while not self._timed_out:
      if self._check_for_ok():
        return
      else:
        sleep(self.sleep_interval)
    else:
      raise SudoTimeout('Timeout awaiting sudo ok (incorrect password?)')

  def prepare(self, argv, cwd=None):
    if self.ok_path is not None:
      self.cleanup()
    self.temp_dir = mkdtemp()
    self.ok_path = os.path.join(self.temp_dir, 'ok')
    argv = self._update_argv(argv, cwd)
    if self.password:
      self.fifo_path = os.path.join(self.temp_dir, 'fifo')
      self.askpass_path = os.path.join(self.temp_dir, 'askpass')
      self._fifo_create()
      self._askpass_create()
    return argv, self.askpass_path

  def wait(self):
    if self.ok_path is None:
      raise ValueError(f'prepare() must be called before wait()')
    try:
      if self._check_for_ok():
        return
      self.timer = Timer(start=True)
      if self.password:
        self._handle_fifo()
      self._await_ok()
    finally:
      self.cleanup()

  def cleanup(self):
    if self.temp_dir and os.path.isdir(self.temp_dir):
      shutil.rmtree(self.temp_dir)
    self.timer = None
    self.temp_dir = None
    self.fifo_path = None
    self.ok_path = None
    self.askpass_path = None

SudoHelper._askpass_tmpl = '''#!{{ quote(shell) }}
cat < {{ quote(fifo_path) }}
'''

def popen(
  argv, *args, sudo_user=None, sudo_group=None, sudo_password=None,
  sudo_login=None, env=None, cwd=None, **kwargs
):
  if env is None:
    env = dict(os.environ)
  sudo_helper = SudoHelper(
    user=sudo_user, group=sudo_group, password=sudo_password, login=sudo_login)
  argv, askpass_path = sudo_helper.prepare(argv, cwd)
  try:
    if askpass_path:
      env['SUDO_ASKPASS'] = askpass_path
    proc = subp.Popen(argv, *args, env=env, **kwargs)
    sudo_helper.wait()
    return proc
  finally:
    sudo_helper.cleanup()



from setuptools import setup, find_packages

name             = 'lura'
version          = '0.4.7'
author           = 'Nick Zigarovich'
author_email     = 'nick@zigarovich.io'
url              = 'https://github.com/ecks0/lura'
description      = 'a bag of tricks'
python_requires  = '>= 3.6'
install_requires = open('requirements.txt').read().strip().splitlines()
packages         = find_packages()
console_scripts  = ['fuga=lura.system.fuga:fuga']
entry_points     = dict(console_scripts=console_scripts)
zip_safe         = True
long_description = open('README.md').read()
long_description_content_type = 'text/markdown'

setup(
  name=name,
  version=version,
  author=author,
  author_email=author_email,
  url=url,
  description=description,
  python_requires=python_requires,
  install_requires=install_requires,
  packages=packages,
  entry_points=entry_points,
  zip_safe=zip_safe,
  long_description=long_description,
  long_description_content_type=long_description_content_type,
)



import io
import logging
import os
import shlex
import subprocess as subp
import sys
import threading
from collections import MutableSequence
from contextlib import contextmanager
from deepmerge import always_merger
from lura import formats
from lura import threads
from lura.attrs import attr
from lura.attrs import ottr
from lura.sudo import SudoHelper
from lura.sudo import shell_path
from lura.utils import deepcopy
from types import GeneratorType

log = logging.getLogger(__name__)

merge = always_merger.merge
shjoin = subp.list2cmdline

def tee(source, targets, cond=lambda: True):
  '`readline` from one source and `write` to many targets.'

  while cond():
    line = source.readline()
    if line == '':
      break
    for target in targets:
      target.write(line)

class Tee(threads.Thread):
  '`readline` from one source and `write` to many targets in a thread.'

  def __init__(self, source, targets):
    super().__init__(name='Tee')
    self.source = source
    self.targets = targets
    self._work = False

  def run(self):
    self._work = True
    tee(self.source, self.targets, lambda: self._work is True)

  def stop(self):
    self._work = False

class LogWriter:
  'A file-like object which logs data written to it.'

  # this is used to implement the `run.log(...)` context manager

  def __init__(self, logger, level, tag):
    super().__init__()
    self.log = logger[level]
    self.tag = tag

  def write(self, line):
    self.log(f'[{self.tag}] {line.rstrip()}')

class Result:
  'The value returned by `run()`.'

  as_dict_members = ('args', 'argv', 'code', 'stdout', 'stderr')

  def __init__(self, context, code, stdout, stderr):
    super().__init__()
    self.args = context.args
    self.argv = context.argv
    self.code = code
    self.stdout = stdout
    self.stderr = stderr
    self.context = context
    del context['args']
    del context['argv']
    if context.sudo_password:
      context.sudo_password = '<scrubbed>'

  def as_dict(self):
    return ottr((_, getattr(self, _)) for _ in self.as_dict_members)

  def format(self, fmt='yaml'):
    tag = 'run.{}'.format(type(self).__name__.lower())
    return formats.for_ext(fmt).dumps({tag: self.as_dict()})

  def print(self, fmt='yaml', file=None):
    file = file or sys.stdout
    file.write(self.format(fmt=fmt))

class Error(RuntimeError):
  'The error raised by `run()` when a process exits with the wrong code.'

  def __init__(self, result):
    msg = 'Process expected exit code {} but received code {}{}{}'.format(
      result.context.enforce_code, result.code, os.linesep, result.format())
    super().__init__(msg)
    self.result = result

class Run(threading.local):
  'The implementation of `run()`.'

  # kwargs that we accept
  _kwargs = (
    'pty', 'env', 'env_replace', 'cwd', 'shell', 'stdin', 'stdout', 'stderr',
    'encoding', 'enforce', 'enforce_code', 'sudo', 'sudo_user', 'sudo_group',
    'sudo_password', 'sudo_login'
  )

  # default kwarg values
  defaults = dict(
    pty           = False,
    env           = None,
    env_replace   = False,
    cwd           = None,
    shell         = False,
    stdin         = None,
    stdout        = [],
    stderr        = [],
    encoding      = sys.getdefaultencoding(),
    enforce       = True,
    enforce_code  = 0,
    sudo          = False,
    sudo_user     = None,
    sudo_group    = None,
    sudo_password = None,
    sudo_login    = False,
  )

  default_shell = shell_path()

  # for stdio threads
  stdio_join_timeout = 0.5

  # when wait()ing on a process, return execution to python at this interval
  interpreter_checkin_interval = 1.0

  def __init__(self):
    super().__init__()
    # context managers set their values here
    self.context = {}

  def _subprocess(self, ctx):
    'Run with `subprocess.Popen`.'

    outbuf, errbuf = io.StringIO(), io.StringIO()
    stdout = [outbuf] + ctx.stdout
    stderr = [errbuf] + ctx.stderr
    argv = ctx.args if ctx.shell else ctx.argv
    proc = None
    threads = ()
    try:
      proc = subp.Popen(
        argv, env=ctx.env, cwd=ctx.cwd, shell=ctx.shell, stdout=subp.PIPE,
        stderr=subp.PIPE, stdin=ctx.stdin, encoding=ctx.encoding)
      if ctx.sudo:
        ctx.sudo_helper.wait()
      threads = (Tee.spawn(proc.stdout, stdout), Tee.spawn(proc.stderr, stderr))
      # it turns out that it's sometimes beneficial to allow execution to
      # return to the interpreter every once in a while. we poll high
      # enough that it shouldn't matter, performance-wise. incidentally,
      # this allows run() to be cancellable when running in a
      # `lura.threads.Thread`.
      while True:
        try:
          code = proc.wait(self.interpreter_checkin_interval)
          break
        except subp.TimeoutExpired:
          continue
      for thread in threads:
        thread.join()
      proc = None
      threads = ()
      return self.Result(ctx, code, outbuf.getvalue(), errbuf.getvalue())
    finally:
      for thread in threads:
        thread.stop()
        thread.join(self.stdio_join_timeout)
        if thread.is_alive():
          log.debug(f'Failed joining tee thread: {thread}')
      outbuf.close()
      errbuf.close()
      if proc:
        proc.kill()

  def _ptyprocess(self, ctx):
    'Run with `PtyProcessUnicode`.'

    if ctx.stdin:
      raise NotImplementedError('stdin not implemented for pty processes')
    if ctx.shell:
      ctx.argv = [self.default_shell, '-c', ctx.args]
      ctx.args = shjoin(ctx.argv)
    outbuf = io.StringIO()
    stdout = [outbuf] + ctx.stdout
    proc = None
    try:
      proc = PtyProcessUnicode.spawn(argv, env=ctx.env, cwd=ctx.cwd)
      if ctx.sudo:
        ctx.sudo_helper.wait()
      reader = attr(readline=lambda: f'{proc.readline()[:-2]}{os.linesep}')
      try:
        tee(reader, stdout)
      except EOFError:
        pass
      code = proc.wait()
      proc = None
      return self.Result(ctx, code, outbuf.getvalue(), '')
    finally:
      outbuf.close()
      if proc:
        try:
          proc.kill(9)
        except Exception:
          log.debug('Failed killing pty process', exc_info=True)

  def _check_args(self, kwargs):
    'Validate arg names in `self.defaults`, `self.context`, and `kwargs`.'

    def check_args(name, bundle):
      for arg in bundle:
        if arg not in self._kwargs:
          raise ValueError(f'Unknown argument in {name}: {arg}={bundle[arg]}')
    check_args('defaults', self.defaults)
    check_args('context', self.context)
    check_args('kwargs', kwargs)

  def _build_context(self, kwargs):
    defaults = deepcopy(self.defaults)
    context = deepcopy(self.context)
    return merge(defaults, merge(context, kwargs))

  def _env_setup(self, ctx):
    'Setup the environment the process will run in.'

    if ctx.env is None:
      ctx.env = {}
    if not ctx.env_replace:
      env = {}
      env.update(os.environ)
      env.update(ctx.env)
      ctx.env = env

  def _sudo_setup(self, ctx, argv):
    'Setup sudo, returning the new sudo argv.'

    ctx.sudo_helper = SudoHelper(
      user=ctx.sudo_user, group=ctx.sudo_group, password=ctx.sudo_password,
      login=ctx.sudo_login)
    argv, askpass_path = ctx.sudo_helper.prepare(argv, ctx.cwd)
    ctx.cwd = None
    if askpass_path:
      ctx.env['SUDO_ASKPASS'] = askpass_path
    return argv

  def _sudo_cleanup(self, ctx):
    ctx.sudo_helper.cleanup()
    del ctx['sudo_helper']

  def _argv_setup(self, ctx, argv):
    'ctx.args is always a string, ctx.argv is always a list.'

    if isinstance(argv, str):
      ctx.args = argv
      ctx.argv = shlex.split(argv)
    else:
      ctx.args = shjoin(argv)
      ctx.argv = argv

  def run(self, argv, **kwargs):
    'Entry point for `run()`.'

    self._check_args(kwargs)
    ctx = self._build_context(kwargs)
    self._env_setup(ctx)
    if ctx.sudo:
      argv = self._sudo_setup(ctx, argv)
    try:
      self._argv_setup(ctx, argv)
      if ctx.pty:
        result = self._ptyprocess(ctx)
      else:
        result = self._subprocess(ctx)
      if ctx.enforce and result.code != ctx.enforce_code:
        raise self.Error(result)
      return result
    finally:
      if ctx.sudo:
        self._sudo_cleanup(ctx)

  __call__ = run

  def zero(*args, **kwargs):
    '''
    A predicate which returns `True` if `argv` exits with code 0, else `False`.

    Sets `enforce` to `False` if not present in `kwargs`.
    '''

    kwargs.setdefault('enforce', False)
    return self.run(*args, **kwargs).code == 0

  def nonzero(*args, **kwargs):
    '''
    A predicate which returns `True` if `argv` exits with a non-zero exit
    code, else `True`.

    Sets `enforce` to `False` if not present in `kwargs`.
    '''

    kwargs.setdefault('enforce', False)
    return self.run(*args, **kwargs).code != 0

  @contextmanager
  def quash(self):
    'Do not enforce exit code while in this context.'

    old = self.context
    new = {'enforce': False}
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def enforce(self, enforce_code=0):
    'Enforce exit code `enforce_code` while in this context.'

    old = self.context
    new = {'enforce': True, 'enforce_code': enforce_code}
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def cwd(self, cwd):
    'Run in directory `cwd` while in this context.'

    old = self.context
    new = {'cwd': cwd}
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def shell(self):
    "Run all commands with the user's shell while in this context."

    old = self.context
    new = {'shell': True}
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def sudo(
    self, sudo_user=None, sudo_group=None, sudo_password=None, sudo_login=None
  ):
    'Run with sudo while in this context.'

    old = self.context
    new = dict(
      sudo = True,
      sudo_user = sudo_user,
      sudo_group = sudo_group,
      sudo_password = sudo_password,
      sudo_login = sudo_login,
    )
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def std(self, stdout=None, stderr=None):
    '''
    Send stdout and/or stderr to the given file objects or lists of file
    objects while in this context.
    '''

    if stdout:
      if isinstance(stdout, (Sequence, GeneratorType)):
        stdout = list(stdout)
      else:
        stdout = [stdout]
    else:
      stdout = []
    if stderr:
      if isinstance(stderr, (Sequence, GeneratorType)):
        stderr = list(stderr)
      else:
        stderr = [stderr]
    else:
      stderr = []
    old = self.context
    new = {'stdout': stdout, 'stderr': stderr}
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

  @contextmanager
  def log(self, logger, log_level=log.DEBUG):
    'Send stdout and stderr to a logger while in this context.'

    old = self.context
    new = {
      'stdout': [LogWriter(logger, log_level, 'out')],
      'stderr': [LogWriter(logger, log_level, 'err')],
    }
    self.context = merge(deepcopy(old), new)
    try:
      yield
    finally:
      self.context = old

Run.Error = Error
Run.Result = Result

run = Run()



'Ssh client with sudo support.'

import fabric
import logging
import os
import sys
from invoke import Responder
from shlex import quote
from subprocess import list2cmdline as shjoin

logger = logging.getLogger(__name__)

class Client:

  log_level = logger.DEBUG

  _default_path = ':'.join((
    '/bin',
    '/usr/bin',
    '/usr/local/bin',
    '/sbin',
    '/usr/sbin',
    '/usr/local/sbin',
    '/usr/libexec/'
  ))

  def __init__(
    self, host, port=22, user=None, password=None, key_file=None,
    private_key=None, passphrase=None, connect_timeout=60.0, auth_timeout=60.0,
    sudo_password=None, compress=True, path=None
  ):
    # FIXME accept key data from buffer
    super().__init__()
    self._host = host
    self._port = port
    self._user = user
    self._connect_timeout = connect_timeout
    self._conn_kwargs = {
      'key_filename': key_file,
      'pkey': private_key,
      'passphrase': passphrase,
      'auth_timeout': auth_timeout,
      'compress': compress,
    }
    self._sudo_password = sudo_password
    self._conn = None
    self._path = path if path else self._default_path

  def __del__(self):
    self.close()

  def __enter__(self):
    self.connect()
    return self

  def __exit__(self, *exc_info):
    self.close()

  def _connect(self):
    log = logger[self.log_level]
    log(f'[{self._host}] Connecting')
    overrides = {}
    if self._sudo_password:
      overrides['sudo'] = {'password': self._sudo_password}
    config = fabric.Config(overrides=overrides)
    self._conn = fabric.Connection(
      host=self._host, user=self._user, port=self._port,
      connect_timeout=self._connect_timeout, connect_kwargs=self._conn_kwargs,
      config=config)
    log(f'[{self._host}] Connected')

  def connect(self):
    if self._conn is None:
      self._connect()

  def is_connected(self):
    return self._conn is not None

  def is_closed(self):
    return self._conn is None

  def _close(self):
    log = logger[self.log_level]
    log(f'[{self._host}] Closing')
    try:
      self._conn.close()
    finally:
      self._conn = None
      log(f'[{self._host}] Closed')
      self._host = None

  def close(self):
    if self._conn is not None:
      self._close()

  def put(self, src, dst):
    log = logger[self.log_level]
    self.connect()
    msg = os.linesep.join([
      f'[{self._host}] put:',
      f'[{self._host}]   src: {src}',
      f'[{self._host}]   dst: {dst}',
    ])
    log(msg)
    self._conn.put(src, remote=dst)

  def get(self, src, dst):
    log = logger[self.log_level]
    self.connect()
    msg = os.linesep.join([
      f'[{self._host}] get:',
      f'[{self._host}]   src: {src}',
      f'[{self._host}]   dst: {dst}',
    ])
    log(msg)
    self._conn.get(src, local=dst)

  def _argv(self, argv, cwd):
    if not isinstance(argv, str):
      argv = shjoin(argv)
    user_argv = argv
    argv = [f'export PATH={self._path}; ']
    if cwd:
      argv.append(f'cd {quote(cwd)} && ')
    argv.append(user_argv)
    return f"bash -c {quote(''.join(argv))}"

  def run(
    self, argv, shell=False, pty=False, env={}, replace_env=False,
    encoding=None, stdin=None, stdout=None, stderr=None, enforce=True,
    cwd=None
  ):
    log = logger[self.log_level]
    self.connect()
    argv = self._argv(argv, cwd)
    log(f'[{self._host}] run: {argv}')
    return self._conn.run(
      argv, shell=shell, pty=pty, env=env, replace_env=replace_env,
      encoding=encoding, in_stream=stdin, out_stream=stdout,
      err_stream=stderr, warn=not enforce, hide=True)

  def sudo(
    self, argv, shell=False, pty=False, env={}, replace_env=False,
    encoding=None, stdin=None, stdout=None, stderr=None, enforce=True,
    user=None, login=False, cwd=None
  ):
    log = logger[self.log_level]
    self.connect()
    user_argv = self._argv(argv, cwd)
    argv = ['sudo']
    if login:
      argv.append('-i')
    if user:
      argv.extend('-u', user)
    argv.append('--')
    argv.append(user_argv)
    argv = ' '.join(argv)
    log(f'[{self._host}] sudo: {argv}')
    return self._conn.sudo(
      argv, shell=shell, pty=pty, env=env, replace_env=replace_env,
      encoding=encoding, in_stream=stdin, out_stream=stdout,
      err_stream=stderr, warn=not enforce, hide=True)

  def forward(self, lport, rport, lhost=None):
    self.connect()
    return self._conn.forward_local(lport, rport, lhost)

  connected = property(is_connected)
  closed = property(is_closed)



import logging
import os
import sys
import shutil
import tempfile
from time import sleep

log = logging.getLogger(__name__)

def load(path):
  with open(path, 'rb') as pathf:
    return pathf.read()

def loads(path, encoding=None):
  encoding = encoding or sys.getdefaultencoding()
  with open(path, 'r', encoding=encoding) as pathf:
    return pathf.read()

def dump(path, data):
  with open(path, 'wb') as pathf:
    pathf.write(data)

def dumps(path, data, encoding=None):
  encoding = encoding or sys.getdefaultencoding()
  with open(path, 'w', encoding=encoding) as pathf:
    pathf.write(data)

def append(path, data):
  with open(path, 'ab') as pathf:
    pathf.write(data)

def appends(path, data, encoding=None):
  encoding = encoding or sys.getdefaultencoding()
  with open(path, 'a', encoding=encoding) as pathf:
    patf.write(data)

def follow(path, interval=0.5, cond=lambda: True):
  with open(path) as pathf:
    buf = []
    while cond():
      data = pathf.readline()
      if data == '':
        sleep(interval)
        continue
      if not data.endswith(os.linesep):
        buf.append(data)
        continue
      yield ''.join(buf) + data

class TempDir:

  def __init__(self, suffix=None, prefix=None, dir=None, keep=False):
    super().__init__()
    if prefix is None:
      prefix = 'lura.'
    self._tempdir_suffix = suffix
    self._tempdir_prefix = prefix
    self._tempdir_root = dir
    self._tempdir_keep = keep
    self._tempdir_dir = None

  def __enter__(self):
    self._tempdir_dir = tempfile.mkdtemp(
      suffix=self._tempdir_suffix, prefix=self._tempdir_prefix,
      dir=self._tempdir_root)
    return self._tempdir_dir

  def __exit__(self, *exc_info):
    if self._tempdir_keep:
      log.warn(f'Keeping temporary directory: {self._tempdir_dir}')
    else:
      shutil.rmtree(self._tempdir_dir)
    self._tempdir_dir = None

class TempFile(TempDir):

  def __init__(self, *args, **kwargs):
    super().__init__()

  def __enter__(self):
    temp_dir = super().__enter__()
    return os.path.join(temp_dir, 'file')
